@@ -2,7 +2,7 @@
 
 ;; Copyright (C) 2018 Yuan Fu
 
-;; Version: 1.11.1
+;; Version: 1.12.1
 
 ;; Author: Yuan Fu <casouri@gmail.com>
 ;; URL: https://github.com/casouri/eldoc-box
@@ -84,6 +84,10 @@
 (defface eldoc-box-body '((t . nil))
   "Body face used in documentation childframe.")
 
+(defface eldoc-box-markdown-separator '((t . ( :strike-through t
+                                               :height 0.4)))
+  "Face for the separator line in Markdown.")
+
 (defcustom eldoc-box-lighter " ELDOC-BOX"
   "Mode line lighter for all eldoc-box modes.
 If the value is nil, no lighter is displayed."
@@ -146,7 +150,8 @@ This separator is used for the documentation shown in
     (drag-internal-border . t)
     (no-special-glyphs . t)
     (desktop-dont-save . t)
-    (tab-bar-lines . 0))
+    (tab-bar-lines . 0)
+    (tab-bar-lines-keep-state . 1))
   "Frame parameters used to create the frame.")
 
 (defcustom eldoc-box-max-pixel-width 800
@@ -173,8 +178,17 @@ Its value should be a list: (left right top)"
 
 (defvar eldoc-box-position-function #'eldoc-box--default-upper-corner-position-function
   "Eldoc-box uses this function to set childframe's position.
-This should be a function that returns a (X . Y) cons cell.
-It will be passes with two arguments: WIDTH and HEIGHT of the childframe.")
+
+The function is passed two arguments, WIDTH and HEIGHT of the
+childframe, and should return a (X . Y) cons cell.")
+
+(defvar eldoc-box-at-point-position-function #'eldoc-box--default-at-point-position-function
+  "Eldoc-box uses this function to set childframe's position.
+This function is used in ‘eldoc-box-help-at-point’ and in
+‘eldoc-box-hover-at-point-mode’.
+
+The function is passed two arguments, WIDTH and HEIGHT of the
+childframe, and should return a (X . Y) cons cell.")
 
 (defcustom eldoc-box-fringe-use-same-bg t
   "T means fringe's background color is set to as same as that of default."
@@ -212,6 +226,10 @@ See `eldoc-box-inhibit-display-when-moving'."
   (when (and eldoc-box--frame (frame-live-p eldoc-box--frame))
     (make-frame-invisible eldoc-box--frame t)))
 
+(defvar-local eldoc-box--old-eldoc-functions nil
+  "The original value of ‘eldoc-display-functions’.
+The original value before enabling eldoc-box.")
+
 (defun eldoc-box--enable ()
   "Enable eldoc-box hover.
 Intended for internal use."
@@ -219,6 +237,8 @@ Intended for internal use."
       (add-function :before-while (local 'eldoc-message-function)
                     #'eldoc-box--eldoc-message-function)
 
+    (setq-local eldoc-box--old-eldoc-functions
+                eldoc-display-functions)
     (setq-local eldoc-display-functions
                 (cons 'eldoc-box--eldoc-display-function
                       (remq 'eldoc-display-in-echo-area
@@ -234,12 +254,18 @@ Intended for internal use."
       (remove-function (local 'eldoc-message-function) #'eldoc-box--eldoc-message-function)
 
     (setq-local eldoc-display-functions
-                (cons 'eldoc-display-in-echo-area
-                      (remq 'eldoc-box--eldoc-display-function
-                            eldoc-display-functions))))
+                (remq 'eldoc-box--eldoc-display-function
+                      eldoc-display-functions))
+    ;; If we removed eldoc-display-in-echo-area when enabling
+    ;; eldoc-box, add it back.
+    (when (memq 'eldoc-display-in-echo-area
+                eldoc-box--old-eldoc-functions)
+      (setq-local eldoc-display-functions
+                  (cons 'eldoc-display-in-echo-area
+                        eldoc-display-functions))))
 
   (advice-remove #'keyboard-quit #'eldoc-box-quit-frame)
-  ;; If minor mode is turned off when childframe is visible, hide it.
+  ;; If minor mode is turned off when the childframe is visible, hide it.
   (when eldoc-box--frame
     (delete-frame eldoc-box--frame)
     (setq eldoc-box--frame nil)))
@@ -247,30 +273,52 @@ Intended for internal use."
 ;;;;; Help at point
 
 (defvar eldoc-box--help-at-point-last-point 0
-  "This point cache is used by clean up function.
-If (point) != last point, cleanup frame.")
+  "This point cache is used by the clean up function.
+If point != last point, hide the childframe.")
 
 (defun eldoc-box--help-at-point-cleanup ()
-  "Try to clean up the childframe made by eldoc-box hack."
+  "Try to clean up the childframe."
   (if (or (eq (point) eldoc-box--help-at-point-last-point)
-          ;; don't clean up when the user clicks childframe
+          ;; Don't clean up when the user clicks into the childframe.
           (eq (selected-frame) eldoc-box--frame))
       (run-with-timer 0.1 nil #'eldoc-box--help-at-point-cleanup)
-    (eldoc-box-quit-frame)
-    (kill-local-variable 'eldoc-display-functions)))
+    (eldoc-box-quit-frame)))
+
+(defun eldoc-box--help-at-point-async-update (docs _interactive)
+  "Update async doc changes to help-at-point childframe.
+
+This is added to ‘eldoc-display-functions’, such that when async doc
+comes in, the at-point doc pop-up can be updated.
+
+For DOCS, see ‘eldoc-display-functions’."
+  (when (and (frame-visible-p eldoc-box--frame)
+             (eq eldoc-box--help-at-point-last-point (point)))
+    (let ((eldoc-box-position-function
+           eldoc-box-at-point-position-function))
+      (eldoc-box--display
+       (concat (mapcar #'car docs)
+               (concat "\n"
+                       (or eldoc-doc-buffer-separator "---")
+                       "\n"))))))
 
 ;;;###autoload
 (defun eldoc-box-help-at-point ()
   "Display documentation of the symbol at point."
   (interactive)
   (when (boundp 'eldoc--doc-buffer)
+    (add-hook 'eldoc-display-functions
+              #'eldoc-box--help-at-point-async-update 0 t)
     (let ((eldoc-box-position-function
-           #'eldoc-box--default-at-point-position-function))
+           eldoc-box-at-point-position-function)
+          (doc (with-current-buffer eldoc--doc-buffer
+                 (buffer-string))))
       (eldoc-box--display
-       (with-current-buffer eldoc--doc-buffer
-         (buffer-string))))
+       (if (equal doc "")
+           "There’s no doc to display at this point" doc)))
     (setq eldoc-box--help-at-point-last-point (point))
-    (run-with-timer 0.1 nil #'eldoc-box--help-at-point-cleanup)))
+    (run-with-timer 0.1 nil #'eldoc-box--help-at-point-cleanup)
+    (when eldoc-box-clear-with-C-g
+      (advice-add #'keyboard-quit :before #'eldoc-box-quit-frame))))
 
 ;;;; Backstage
 ;;;;; Variable
@@ -281,6 +329,7 @@ If (point) != last point, cleanup frame.")
 
 ;; Please compiler.
 (defvar eldoc-box-hover-mode)
+
 (defun eldoc-box--display (str)
   "Display STR in childframe.
 STR has to be a proper documentation, not empty string, not nil, etc."
@@ -304,15 +353,24 @@ STR has to be a proper documentation, not empty string, not nil, etc."
       (run-hook-with-args 'eldoc-box-buffer-hook))
     (eldoc-box--get-frame doc-buffer)))
 
-
 (defun eldoc-box--window-side ()
   "Return the side of the selected window.
 Symbol ‘left’ if the selected window is on the left, ‘right’ if
 on the right. Return ‘left’ if there is only one window."
-  (let ((left-window (if tab-bar-mode
-                         (window-at-x-y 5 (+ 5 (tab-bar-height nil t)))
-                       (window-at 0 0))))
-    (if (eq left-window (selected-window))
+  ;; Calculate the left and right distances to the frame edge of the
+  ;; active window.  If the left distance is less than or equal to the
+  ;; right distance, it indicates that the active window is on the left.
+  ;; Otherwise, it is on the right.
+  (let* ((window-left (nth 0 (window-absolute-pixel-edges)))
+         (window-right (nth 2 (window-absolute-pixel-edges)))
+         (frame-left (nth 0 (frame-edges)))
+         (frame-right (nth 2 (frame-edges)))
+         (distance-left (- window-left frame-left))
+         (distance-right (- frame-right window-right)))
+    ;; When `distance-left' equals `distance-right', it means there is
+    ;; only one window in current frame, or the current active window
+    ;; occupies the entire frame horizontally, return left.
+    (if (<= distance-left distance-right)
         'left
       'right)))
 
@@ -334,6 +392,9 @@ Position is calculated base on WIDTH and HEIGHT of childframe text window"
 The coordinate is relative to the native frame.
 
 WINDOW nil means use selected window."
+  (unless point
+    ;; Handle edge case. See https://debbugs.gnu.org/cgi/bugreport.cgi?bug=69259.
+    (setq point (window-point window)))
   (let* ((pos (pos-visible-in-window-p point window t))
          (x (car pos))
          (en (frame-char-width))
@@ -373,6 +434,9 @@ base on WIDTH and HEIGHT of childframe text window."
          (y (cdr pos)))
     (cons (or (eldoc-box--at-point-x-by-company) x)
           y)))
+
+(defvar eldoc-box--markdown-separator-display-props)
+
 (defun eldoc-box--update-childframe-geometry (frame window)
   "Update the size and the position of childframe.
 FRAME is the childframe, WINDOW is the primary window."
@@ -380,13 +444,18 @@ FRAME is the childframe, WINDOW is the primary window."
   ;; property of (space :width text) -- which is what we apply onto
   ;; markdown separators -- ‘window-text-pixel-size’ wouldn’t return
   ;; the correct value. Instead, it returns the current window width.
-  ;; So now the childram only grows in size and never shrinks. For
-  ;; whatever reason, if we set the frame size very small before
-  ;; calculating window’s text size, it can return the right value.
+  ;; So now the childram only grows in size and never shrinks.
+  ;;
   ;; (My guess is that the function takes (space :width text) at face
   ;; value, but that can’t be the whole picture because it works fine
   ;; when I manually evaluate the function in the childframe...)
-  (set-frame-size frame 1 1 t)
+  ;;
+  ;; The original workaround of setting the frame size to something
+  ;; small before calling ‘window-text-pixel-size’ works, but brings
+  ;; other problems. Now we just set the display property to nil
+  ;; before calling ‘window-text-pixel-size’, and set them back after.
+  (setcdr eldoc-box--markdown-separator-display-props nil)
+
   (let* ((size
           (window-text-pixel-size
            window nil nil
@@ -399,6 +468,11 @@ FRAME is the childframe, WINDOW is the primary window."
          (frame-resize-pixelwise t)
          (pos (funcall eldoc-box-position-function width height)))
     (set-frame-size frame width height t)
+
+    ;; Set the display property back.
+    (setcdr eldoc-box--markdown-separator-display-props
+            '(:width text))
+
     ;; move position
     (set-frame-position frame (car pos) (cdr pos))))
 
@@ -408,8 +482,8 @@ FRAME is the childframe, WINDOW is the primary window."
     (setq eldoc-box--inhibit-childframe t)
     (eldoc-box-quit-frame)
     (run-with-idle-timer sec nil
-                    (lambda ()
-                      (setq eldoc-box--inhibit-childframe nil)))))
+                         (lambda ()
+                           (setq eldoc-box--inhibit-childframe nil)))))
 
 (defun eldoc-box--follow-cursor ()
   "Make childframe follow cursor in at-point mode."
@@ -454,16 +528,14 @@ Checkout `lsp-ui-doc--make-frame', `lsp-ui-doc--move-frame'."
       (set-face-attribute 'internal-border frame :inherit 'eldoc-box-border)
       (when (facep 'child-frame-border)
         (set-face-background 'child-frame-border
-                             (face-attribute 'eldoc-box-border :background)
+                             (face-attribute 'eldoc-box-border :background nil t)
                              frame))
-      ;; set size
       (eldoc-box--update-childframe-geometry frame window)
       (setq eldoc-box--frame frame)
       (with-selected-frame frame
         (run-hook-with-args 'eldoc-box-frame-hook main-frame))
       (make-frame-visible frame))))
 
-
 ;;;;; ElDoc
 
 (defvar eldoc-box--cleanup-timer nil
@@ -592,7 +664,7 @@ display the docs in echo area depending on
 
 ;;;###autoload
 (define-minor-mode eldoc-box-hover-mode
-  "Displays hover documentations in a childframe.
+  "Display hover documentations in a childframe.
 The default position of childframe is upper corner."
   :lighter eldoc-box-lighter
   (if eldoc-box-hover-mode
@@ -604,13 +676,13 @@ The default position of childframe is upper corner."
 ;;;###autoload
 (define-minor-mode eldoc-box-hover-at-point-mode
   "A convenient minor mode to display doc at point.
-You can use \[keyboard-quit] to hide the doc."
+You can use \\[keyboard-quit] to hide the doc."
   :lighter eldoc-box-lighter
   (if eldoc-box-hover-at-point-mode
       (progn (when eldoc-box-hover-mode
                (eldoc-box-hover-mode -1))
              (setq-local eldoc-box-position-function
-                         #'eldoc-box--default-at-point-position-function)
+                         eldoc-box-at-point-position-function)
              (setq-local  eldoc-box-clear-with-C-g t)
              (remove-hook 'pre-command-hook #'eldoc-pre-command-refresh-echo-area t)
              (add-hook 'post-command-hook #'eldoc-box--follow-cursor t t)
@@ -633,25 +705,40 @@ instead."
   (interactive)
   (eldoc-box-help-at-point))
 
-;;;; Comany compatibility
+;;;; Company compatibility
 ;;
 
 ;; see also `eldoc-box--default-at-point-position-function'
 
 ;; please compiler
 (defvar company-pseudo-tooltip-overlay)
+(declare-function company-box--get-frame "company-box")
 
 (defun eldoc-box--at-point-x-by-company ()
   "Return the x position that accommodates company's popup."
-  (if (and (featurep 'company) company-pseudo-tooltip-overlay)
-      (+ (* (frame-char-width)
-            (+ (overlay-get company-pseudo-tooltip-overlay 'company-width)
-               (overlay-get company-pseudo-tooltip-overlay 'company-column)))
-         (or (line-number-display-width t) 0))
-    nil))
+  (cond
+   ((and (boundp 'company-pseudo-tooltip-overlay)
+         company-pseudo-tooltip-overlay)
+    (+ (* (frame-char-width)
+          (+ (overlay-get company-pseudo-tooltip-overlay
+                          'company-width)
+             (overlay-get company-pseudo-tooltip-overlay
+                          'company-column)))
+       (or (line-number-display-width t) 0)))
+   ((and (boundp 'company-box--x) (numberp company-box--x))
+	(+ company-box--x
+       (frame-pixel-width (company-box--get-frame))))
+   (t nil)))
 
 ;;;; Markdown compatibility
 
+(defvar-local eldoc-box--markdown-separator-display-props
+    '(space :width text)
+  "Stores the display text property applied to markdown separators.
+
+Due to a bug, in ‘eldoc-box--update-childframe-geometry’, we
+modify the display property temporarily and then set it back.")
+
 (defun eldoc-box--prettify-markdown-separator ()
   "Prettify the markdown separator in doc returned by Eglot.
 Refontify the separator so they span exactly the width of the
@@ -660,11 +747,11 @@ childframe."
     (goto-char (point-min))
     (let (prop)
       (while (setq prop (text-property-search-forward 'markdown-hr))
-        (add-text-properties (prop-match-beginning prop)
-                             (prop-match-end prop)
-                             '( display (space :width text)
-                                face ( :strike-through t
-                                       :height 0.4)))))))
+        (add-text-properties
+         (prop-match-beginning prop)
+         (prop-match-end prop)
+         `( display ,eldoc-box--markdown-separator-display-props
+            face eldoc-box-markdown-separator))))))
 
 (defun eldoc-box--replace-en-space ()
   "Display the en spaces in documentation as regular spaces."
@@ -731,6 +818,17 @@ height."
                          'invisible t)
       (put-text-property (match-beginning 3) (match-end 3)
                          'invisible t))
+    ;; Don't show these tags.
+    (goto-char (point-min))
+    (while (re-search-forward
+            (rx (group "<p>")
+                (group (*? anychar))
+                (group "</p>"))
+            nil t)
+      (put-text-property (match-beginning 1) (match-end 1)
+                         'invisible t)
+      (put-text-property (match-beginning 3) (match-end 3)
+                         'invisible t))
     ;; Special entities.
     (goto-char (point-min))
     (while (re-search-forward (rx (or "&lt;" "&gt;" "&nbsp;")) nil t)
@@ -747,7 +845,9 @@ height."
   "Discard the current childframe and regenerate one.
 This allows any change in childframe parameter to take effect."
   (interactive)
-  (setq eldoc-box--frame nil))
+  (when eldoc-box--frame
+    (delete-frame eldoc-box--frame)
+    (setq eldoc-box--frame nil)))
 
 (with-eval-after-load 'tab-bar
   (add-hook 'tab-bar-mode-hook #'eldoc-box-reset-frame))
